package hu.stunts.scr.apps;

import hu.stunts.scr.exceptions.InvalidReplayFileException;
import hu.stunts.scr.exceptions.MissingReferenceTrackException;
import hu.stunts.scr.utils.*;
import hu.stunts.scr.utils.enums.*;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.Map;


public abstract class BaseReplay implements Replay, Comparable<Replay> {

	protected static final int MAXINFOFIELDS = 12; // No. of fields to be extracted from the file
	private static final int RPLBSIZE = 13848; // size of a 10-min. replay
	private static final int TRKBSIZE = 1802; // size of a track
	private double timeEpsilon = 0.005; // Minimum significant time difference between laps, in seconds
	private int frameRate = 20; // Frames per second
	private File rplFile;
	private boolean isValid = false; // checkIntegrity() needs to be used only once!
	private String trackName = "";
	private String carID = "";
	private String racer = "";
	private boolean isPenaltyMatched = false;
	private int penaltyTime = 0;
	private Map<String, String> penaltyDict = null;
	private boolean isDQ = false; //unimplemented
	
	public BaseReplay(String filePath, Map<String, String>... dictionaries) {
		try {
			rplFile = new File (filePath);
			if (checkIntegrity()){
				isValid = true;
				racer = rplFile.getName().substring(0,rplFile.getName().length()-4); //decoupling from the set-method
				readCarID();
				readTrackName();
				if (dictionaries[BaseReplayDicts.PENALTY] != null)
					penaltyDict = dictionaries[BaseReplayDicts.PENALTY];
				else
					initializePenaltyDict();
				isPenaltyMatched = readPenaltyDefault();
			}
		} catch (InvalidReplayFileException e) {
			System.err.println (filePath+" is not a valid .RPL!");
			e.printStackTrace();
		} 
	}
	
	public BaseReplay(String filePath) {
		this(filePath, (Map<String, String>)null);
	}
	
	//base class constructor methods
	
	private boolean checkIntegrity() throws InvalidReplayFileException {
		boolean answer = false;
		RandomAccessFile rplStream = null;
		try {
			int keyEventLength = 0;
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			if (newFormat())
				answer = true;
			else {
				rplStream.seek(0x16);
				keyEventLength = IOStunts.readStuntsUShort(rplStream);
				if ((keyEventLength + 0x722) == rplFile.length())
					answer = true;
				else {
					throw new InvalidReplayFileException();
				}
				rplStream.close();
			}
		} catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
			throw new InvalidReplayFileException();
		}
		catch (IOException e) {
			System.err.println ("Validity checking failed!");
			e.printStackTrace();
			throw new InvalidReplayFileException();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return answer;
	}

	private void readCarID () {
		RandomAccessFile rplStream = null;
		try {
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			rplStream.seek(0x0);
			carID = IOStunts.readStuntsString(rplStream, 4);
		} catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("Could not read car ID!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
	}

	private void readTrackName () {
		RandomAccessFile rplStream = null;
		try {
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			rplStream.seek(0x0d);
			trackName = IOStunts.readStuntsString(rplStream, 9);
			rplStream.close();
		}	catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("Could not read track name!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
	}
	
	/* On the subject of penalty dictionary boot, note that penaltyDict will
	 * end up as null if penalty.txt is missing.
	 */
	private void initializePenaltyDict() {
		int dictInitSize = 50;
		File penaltyASCII = new File(rplFile.getParent() + File.separator + "penalty.txt");
		penaltyDict = TxtParsing.simpleMakeMapFromPairs(penaltyASCII, dictInitSize);
	}
	
	private boolean readPenaltyDefault () {
		boolean found = false;
		double pen = 0;
		if (penaltyDict != null) { //if the dict was properly initialized
			String val = penaltyDict.get(rplFile.getName());
			if (val != null) {
				pen = Double.parseDouble(val);
				pen = pen * frameRate;
				found = true;
			}
		}
		penaltyTime = (int)Math.round(pen);
		return found;
	}
	
	//public set methods

	public void setRacerName (String name){
		racer = name;
	}

	public void setDQ (boolean flag){
		isDQ = flag;
	}
	
	/* This method is meant to be used with an external penalty database,
	 * passed to the constructor, which gets updated during the lifetime
	 * of the replay.
	 */
	public void refreshPenalty (){
		isPenaltyMatched = readPenaltyDefault();
	}
	
	//low level IO query methods
	
	public int colourIndex() {
		int index = 0;
		RandomAccessFile rplStream = null;
		try {
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			rplStream.seek(0x04);
			index = rplStream.readByte();
			rplStream.close();
		}	catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("Could not read colour!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return index;
	}
	
	public boolean autoTransmission() {
		boolean answer = true;
		RandomAccessFile rplStream = null;
		try {
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			rplStream.seek(0x05);
			if ((rplStream.readByte() | 0) == 0)
				answer = false;
			rplStream.close();
		}	catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("Could not read transmission!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return answer;
	}
	
	public int opponentIndex() {
		int index = 0;
		RandomAccessFile rplStream = null;
		try {
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			rplStream.seek(0x06);
			index = (rplStream.readByte());
			rplStream.close();
		}	catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("Could not read opponent index!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return index;
	}
	
	public String opponentCarID() {
		String oppCarID = "";
		RandomAccessFile rplStream = null;
		try {
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			rplStream.seek(0x07);
			oppCarID = IOStunts.readStuntsString(rplStream, 4);
		} catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("Could not read car ID!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return oppCarID;
	}
	
	public final boolean newFormat() {
		boolean answer = false;
		RandomAccessFile rplStream = null;
		try {
			int keyEventLength = 0;
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			rplStream.seek(0x18);
			keyEventLength = IOStunts.readStuntsUShort(rplStream);
			if ((keyEventLength + 0x724) == rplFile.length())
				answer = true;
		} catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		}
		catch (IOException e) {
			System.err.println ("Could not read keyboard event blocks!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return answer;
	}
	
	public final boolean checkTermination() {
		boolean answer = true;
		RandomAccessFile rplStream = null;
		try {
			rplStream = new RandomAccessFile (rplFile.getPath(),"r");
			int i;
			long currentPosition = rplFile.length()-1;
			for (i = 0; i < 20; i++){
				rplStream.seek(currentPosition);
				if (rplStream.readByte() != 0)
					answer = false;
				currentPosition--;
			}
			rplStream.close();
		} catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("Could not read final bytes!");
			e.printStackTrace();
		} finally {
			if (rplStream != null)
				try {
					rplStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return answer;
	}

	private File alternateTrkFile(File trkFile) {
		File alternate = trkFile;
		if (!alternate.exists()) {
			String name = trkFile.getName().substring(0,trkFile.getName().length() - 4);
			String extension = trkFile.getName().substring(trkFile.getName().length() - 4, trkFile.getName().length());
			alternate = new File(trkFile.getParentFile(), name + extension.toLowerCase());
			if (!alternate.exists()) {
				alternate = new File(trkFile.getParentFile(), name.toLowerCase() + extension.toLowerCase());
				if (!alternate.exists()) {
					alternate = new File(trkFile.getParentFile(), name.toLowerCase() + extension);
					if (!alternate.exists()) {
						alternate = new File(trkFile.getParentFile(), name.substring(0,1) + name.toLowerCase().substring(1) + extension);
						if (!alternate.exists()) {
							alternate = new File(trkFile.getParentFile(), name.substring(0,1) + name.toLowerCase().substring(1) + extension.toLowerCase());
							if (!alternate.exists()) {
								alternate = trkFile;
							}
						}
					}
				}
			}
		}
		return alternate;
	}
	
	public final boolean checkTrack(File trkFile) throws MissingReferenceTrackException, IOException  {
		int i;
		boolean answer = true;
		FileChannel rplChannel = null, trkChannel = null;
		File altTrkFile = alternateTrkFile(trkFile);
		try {
			if (altTrkFile.exists()){
				rplChannel = new RandomAccessFile (rplFile.getPath(),"r").getChannel();
				trkChannel = new RandomAccessFile (altTrkFile.getPath(),"r").getChannel();
				ByteBuffer rplBB = ByteBuffer.allocate(RPLBSIZE);
				ByteBuffer trkBB = ByteBuffer.allocate(TRKBSIZE);
				rplChannel.read(rplBB);
				rplBB.flip();
				trkChannel.read(trkBB);
				trkBB.flip();
				if (newFormat())
					rplBB.position(0x1a);
				else
					rplBB.position(0x18);
				trkBB.position(0x0);
				for (i = 0; answer != false && i < 0x384; i++) // read up to horizon
					answer = (rplBB.get() == trkBB.get());
				rplBB.position(rplBB.position()+1); // skips horizon
				trkBB.position(trkBB.position()+1);
				for (i = 0; answer != false && i < 0x384; i++)
					answer = (rplBB.get() == trkBB.get());
			}
			else
				throw new MissingReferenceTrackException();
	
		} catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("IO exception on track matching!");
			e.printStackTrace();
		} finally {
			if (rplChannel != null)
				try {
					rplChannel.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			if (trkChannel != null)
				try {
					trkChannel.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return answer;
	}

	public int[] logKeystrokes() {
		int[] logRA = new int[8];
		int keystroke = 0;
		FileChannel rplChannel = null;
		try {
				rplChannel = new RandomAccessFile (rplFile.getPath(),"r").getChannel();
				ByteBuffer rplBB = ByteBuffer.allocate(RPLBSIZE);
				rplChannel.read(rplBB);
				rplBB.flip();
				if (newFormat())
					rplBB.position(0x724);
				else
					rplBB.position(0x722);
				while (rplBB.hasRemaining()) {
					keystroke = rplBB.get();
					if ((keystroke & 1) != 0) logRA[0]++; //accelerate
					if ((keystroke & 2) != 0) logRA[1]++; //brake
					if ((keystroke & 4) != 0) logRA[2]++; //right
					if ((keystroke & 8) != 0) logRA[3]++; //left
					if ((keystroke & 16) != 0) logRA[4]++; //shift up
					if ((keystroke & 32) != 0) logRA[5]++; //shift down
					if ((keystroke & 63) == 0) logRA[6]++; //no input
					if ((keystroke & 64) != 0 ||
							(keystroke & 128) != 0) logRA[7]++; //unknowns
				}
		} catch (FileNotFoundException e) {
			System.err.println ("Replay file not found!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println ("IO exception on keystroke logging!");
			e.printStackTrace();
		} finally {
			if (rplChannel != null)
				try {
					rplChannel.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		return logRA;
	}

	//concrete time computing methods

	public int computeDuration(boolean usePenalty) {
		int time;
		time = (int)rplFile.length() - 1828;
		if (!newFormat())
			time += 2;
		time -= 20;
		if (usePenalty) {
			time = time + penaltyTime;
		}
		return time;
	}
	
	public int computeDuration() {
		return computeDuration(false);
	}
	
	//abstract time computing methods
	
	public abstract double computeTime(boolean usePenalty);
	
	public abstract double computeTime();
	
	public abstract String additionalStateWarnings();
	
	//The comparator works so that slower replays are larger than faster ones.
	public int compareTo(Replay rpl) {
		double diff = (computeTime(true) - rpl.computeTime(true));
		if (Math.abs(diff) < timeEpsilon * frameRate)
			diff = 0;
		return (int)diff;
	}
	
	//info methods and convenience parsers
	public double toSeconds(double rawTime) {
		return rawTime / frameRate;
	}
	
	public double toSeconds(int rawTime) {
		return toSeconds((double)rawTime);
	}
	
	public String[] rplInfo3() {
		String[] output = new String[MAXINFOFIELDS+1];
		output[RI3Fields.WARNINGS] = ""; //error messages
		try {
			if (isValid) {
				if (newFormat())
					output[RI3Fields.VERSION] = "1.1";
				else
					output[RI3Fields.VERSION] = "1.0";
				output[RI3Fields.TRACK] = trackName;
				output[RI3Fields.CAR] = carID;
				output[RI3Fields.COLOUR] = Integer.toString(colourIndex());
				if (autoTransmission())
					output[RI3Fields.TRANSM] = "Auto";
				else
					output[RI3Fields.TRANSM] = "Manual";
				int oppIndex = opponentIndex();
				output[RI3Fields.OPP] = Integer.toString(oppIndex);
				if (oppIndex != 0)
					output[RI3Fields.OPPCAR] = opponentCarID();
				else
					output[RI3Fields.OPPCAR] = "";
				File trkFile = new File (rplFile.getParent()+File.separator+trackName+".TRK");
				try {
					if (checkTrack(trkFile))
						output[RI3Fields.TRKOK] = "Yes";
					else
						output[RI3Fields.TRKOK] = "No";
				} catch (MissingReferenceTrackException e) {
					output[RI3Fields.WARNINGS] = output[RI3Fields.WARNINGS] + "No reference "+trackName+".TRK available.\n";
				} finally {
					if (checkTermination())
						output[RI3Fields.FINISH] = "Yes";
					else
						output[RI3Fields.FINISH] = "No";
					output[RI3Fields.DURATION] = Double.toString(toSeconds(computeDuration()));
					if (penaltyDict != null) {
						output[RI3Fields.TIME] = Double.toString(toSeconds(computeDuration(true))); //Explicitly use base class method
						if (!isPenaltyMatched)
							output[RI3Fields.WARNINGS] = output[RI3Fields.WARNINGS] + "Penalty for this .rpl omitted, assuming 0.00s\n";
					}
					else
						output[RI3Fields.WARNINGS] = output[RI3Fields.WARNINGS] + "Local penalty.txt not found.\n";
				}

			} 
			else 
				output[RI3Fields.WARNINGS] = output[RI3Fields.WARNINGS] + "This is not a valid .RPL!\n";
		} catch (FileNotFoundException e) {
			output[RI3Fields.WARNINGS] = output[RI3Fields.WARNINGS] + "Replay file not found!\n";
			System.err.println ("Replay file not found!");
			//			throw e;
			e.printStackTrace();
		} catch (IOException e) {
			output[RI3Fields.WARNINGS] = output[RI3Fields.WARNINGS] + "IO Exception on RPLInfo!\n";
			System.err.println ("IO exception on RPLInfo!");
			e.printStackTrace();
		}
		return output;
	}
	
	//general-purpose get methods
	
	public String getCar(){
		return carID;
	}

	public String getTrack(){
		return trackName;
	}

	public File getRplFile(){
		return rplFile;
	}

	public String getRacer(){
		return racer;
	}
	
	public int getPenalty() {
		return penaltyTime;
	}
	
	public Map<String, String> getPenaltyDict() {
		return penaltyDict;
	}
	
	public boolean isValid(){
		return isValid;
	}
	
	public boolean isPenaltyMatched(){
		return isValid;
	}

}
