package hu.stunts.scr.apps;

import hu.stunts.scr.exceptions.MissingReferenceTrackException;
import hu.stunts.scr.utils.InfoFormatter;
import hu.stunts.scr.utils.TxtParsing;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.*;

public class FlatDirRace<T extends BaseReplay> implements Race {
	
	private final ClassLoader ccl = Thread.currentThread().getContextClassLoader();
	private URL urlToRacerNames = ccl.getResource("resources/racernames.txt");
	
	private Class<T> rplClass;
	private String raceTitle;
	private File rootDir;
	private RplDirectory rplDir;
	private int keyStart;
	private int keyEnd;
	private File trkFile;
	private boolean performNameLookup = true;
	private InfoFormatter rFormatter = null;
	private Map<String, String> namesDict = null;
	private Map<String, String> bonusDict = null;
	private Map<String, String> penaltyDict = null;
	
	private List<T> activeReplays = new ArrayList<T>();

	private class ucTRKFilter implements FilenameFilter {
		public boolean accept(File dir, String name) {
			return (name.endsWith(".TRK"));
		}
	}
	
	private class MetaRacer extends AbstractInnerRacer<T> {
		private String nameKey = null;
		private List<File> rplList;

		public MetaRacer (List<File> fl) {
			rplList = fl;
		}
		
		public void setNameKey (String key) {
			 nameKey = key;
		}
		
		public T getBestReplay (){
			String racerName = null;
			T currReplay = null;
			T bestReplay = null;
			ListIterator<File> it = rplList.listIterator();
			while (it.hasNext()) {
				File currFile = it.next();
				if (retrieveNameKey(currFile).equals(nameKey)) {
					currReplay = genericReplay(currFile.getPath(), penaltyDict, bonusDict);
					if (currReplay != null && isLegal(currReplay) &&
							(bestReplay == null || (currReplay.compareTo(bestReplay) < 0)))
						bestReplay = currReplay;
					it.remove();
				}
			}
			if (performNameLookup)
				racerName = namesDict.get(nameKey);
			if (!performNameLookup || racerName == null)
				racerName = nameKey;
			if (bestReplay != null)
				bestReplay.setRacerName(racerName);
			return bestReplay;
		}
		
		public String getNameKey () {
			return nameKey;
		}
		
		public T genericReplay (String fileName, Map<String, String>... dictionaries) {
			return genericReplay(rplClass, fileName, dictionaries);
		}
		
		public boolean isLegal (T replay) {
			boolean answer = false;
			try {
				answer = (replay.isValid() &&
						replay.checkTermination());
				if (trkFile != null && trkFile.exists())
					answer = answer && replay.checkTrack(trkFile);
			} catch (MissingReferenceTrackException e) {
				System.err.println("Missing track!");
				e.printStackTrace();
			} catch (IOException e) {
				System.err.println("IO error on track matching!");
				e.printStackTrace();
			}
			return answer;
		} 
	}
	
	public static <T extends BaseReplay> FlatDirRace<T> create(File dir, Class<T> replayClass, int keyStart, int keyEnd, boolean lookName) {
		return new FlatDirRace<T>(dir, replayClass, keyStart, keyEnd, lookName);
	}
	
	public static <T extends BaseReplay> FlatDirRace<T> create(File dir, Class<T> replayClass) {
		return FlatDirRace.create(dir, replayClass, 0, 0, true);
	}
	
	protected FlatDirRace (File dir, Class<T> replayClass, int keyStart, int keyEnd, boolean lookName) {
		rplClass = replayClass;
		rootDir = dir;
		rplDir = new RplDirectory(rootDir);
		rFormatter = new InfoFormatter(dir);
		//track file
		String[] trkList = rootDir.list(new ucTRKFilter());
		if (trkList.length > 0) {
			trkFile = new File(rootDir, trkList[0]);
			raceTitle = trkFile.getName().substring(0, trkFile.getName().length() - 4);
			if (trkList.length > 1) {
				System.err.println("More than one .TRK on root dir, track check disabled.");
				trkFile = null;
			}
		}
		else
			trkFile = null;
		//bonus list
		int dictInitSize = 50;	
		if (rplClass.isAssignableFrom(BonusTableReplay.class)) {
			File bonusASCII = new File(rootDir, "carbonus.txt");
			if (bonusASCII.exists())
				bonusDict = TxtParsing.simpleMakeMapFromPairs(bonusASCII, dictInitSize);
		}
		//penalty list
		File penaltyASCII = new File(rootDir, "penalty.txt");
		if (penaltyASCII.exists())
			penaltyDict = TxtParsing.simpleMakeMapFromPairs(penaltyASCII, dictInitSize);
		//names list
		dictInitSize = 100;
		try {
			InputStream namesIS = urlToRacerNames.openStream();
			namesDict = TxtParsing.simpleMakeMapFromPairs(namesIS, dictInitSize, true);
		} catch (IOException e) {
			e.printStackTrace();
		}
		File extraNamesASCII = new File(rootDir, "racernames.txt");
		if (extraNamesASCII.exists()) {
			Map<String, String> extraNamesDict = TxtParsing.simpleMakeMapFromPairs(extraNamesASCII, dictInitSize, true);
			namesDict.putAll(extraNamesDict);
		}
		this.keyStart = keyStart;
		this.keyEnd = keyEnd;
		performNameLookup = lookName;
	}
	
	private String retrieveNameKey (File f) {
		int begin, end;
		String key = "";
		String fileName = f.getName().substring(0, f.getName().length() - 4);
		begin = keyStart;
		end = keyEnd;
		if (end > fileName.length())
			end = fileName.length();
		if (!(begin < fileName.length()) || end == begin || fileName.length() < end - begin)
			key = fileName;
		else
			key = fileName.substring(begin, end);
		return key.toUpperCase();
	}
	
	public void loadActiveReplays() {
		List<File> rplTempList = new LinkedList<File>();
		activeReplays.clear();
		if (rplDir.getRplCount() > 0)
			for (int i = 0; i < rplDir.getRplCount(); i++)
				rplTempList.add(new File(rootDir, rplDir.getFileName(i)));
		MetaRacer mRacer = new MetaRacer(rplTempList);
		while (!rplTempList.isEmpty()) {
			String key = retrieveNameKey(rplTempList.get(0));
			mRacer.setNameKey(key);
			T activeReplay = mRacer.getBestReplay();
			if (activeReplay != null)
				activeReplays.add(activeReplay);
		}
		Collections.sort(activeReplays);
	}
	
	private class ScoreboardGenerator extends AbstractScoreboardGenerator {
		public ScoreboardGenerator() {
			super(activeReplays, rFormatter);
		}
	}
	
	public String[][] generateHIGStrings (boolean useCorrectedTimes) {
		ScoreboardGenerator sg = new ScoreboardGenerator();
		return sg.generateHIGStrings(useCorrectedTimes);
	}
	
	public void generateHIG (File dest, boolean correctTimes) {
		ScoreboardGenerator sg = new ScoreboardGenerator();
		sg.writeHIG(dest, correctTimes);
	}
	
	public String getRaceTitle() {
		return raceTitle;
	}
	
	public File getRootDir() {
		return rootDir;
	}
	
	public void printTest () {
		for (T replay : activeReplays) {
			System.out.println(replay.getRacer()+" : "+replay.toSeconds(replay.computeTime()));
		}
	}
	
}
