package hu.stunts.scr.ui;

import javax.swing.*;

import java.awt.*;

/**
 * The InfoGridPanel is a JPanel containing a JTextField title element and
 * a two-column pseudo-table (as of now implemented through a GridLayout JPanel
 * containing 'JTextField's). The pseudo-table fields are filled using arrays
 * of strings. Each row has one field plus a header; there are no column
 * headers.
 * The original purpose of this UI element was to display RPLInfo data.
 * @author Daniel David
 * @author Southern Cross Racing Developments
 * @version 0.1
 * @since 0.1
 */

public class InfoGridPanel extends JPanel {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -8267932088596000966L;
	
	private JTextField titleField = new JTextField();
	private JPanel grid = new JPanel();
	private JTextField[] gridHeaders;
	private JTextField[] gridValues;
	
	private int colsHeader;
	private int colsValue;
	
	/**
	 * Standard constructor.
	 * @param title Text to be displayed on the title bar
	 * @param headers Array containing the pseudo-table header contents.
	 * @param values Array containing the pseudo-table value contents.
	 * @param colsH Number of columns of header 'JTextField's
	 * @param colsV Number of columns of value 'JTextField's
	 * @since 0.1
	 */
	public InfoGridPanel (String title, String[] headers, String[] values, int colsH, int colsV){
			colsHeader = colsH;
			colsValue = colsV;
			setLayout(new BoxLayout(this,BoxLayout.Y_AXIS));
			gridHeaders = new JTextField[headers.length];
			gridValues = new JTextField[headers.length];
			setTitleField(title);
			setHeaders(headers);
			setValues(values);
			setGrid();
			add(titleField);
			add(grid);
	}
	
	/**
	 * If omitted, the number of column for 'JTextField's defaults to 20 for
	 * headers and 15 for value fields.
	 * @param title Text to be displayed on the title bar
	 * @param headers Array containing the pseudo-table header contents.
	 * @param values Array containing the pseudo-table value contents.
	 */
	public InfoGridPanel (String title, String[] headers, String[] values) {
		this(title, headers, values, 30, 15);
	}
	
	//acessors and mutators
	
	/**
	 * Defines and formats title bar text.
	 * @param title Text to be displayed on the title bar.
	 */
	public final void setTitleField(String title) {
		titleField.setMaximumSize(new Dimension(320,80));
		titleField.setEditable(false);
		titleField.setText(title);
	}
	
	/**
	 * Sets the layout of the pseudo-table JPanel and adds the previously
	 * created 'JTextField's.
	 */
	public final void setGrid() {
		grid.setLayout(new GridLayout(gridHeaders.length, 2));
		for (int i = 0; i < gridHeaders.length; i++) {
			grid.add(gridHeaders[i]);
			grid.add(gridValues[i]);
		}
	}
	
	/**
	 * Creates the 'JTextField' headers and initialize their contents. 
	 * @param headers Array containing the pseudo-table header contents.
	 */
	public final void setHeaders(String[] headers) {
		for (int i = 0; i < gridHeaders.length; i++) {
			gridHeaders[i] = new JTextField(colsHeader);
			gridHeaders[i].setEditable(false);
			if (i < headers.length)
				gridHeaders[i].setText(headers[i]);
		}
	}
	
	/**
	 * Creates the 'JTextField' value fields and initialize their contents. 
	 * Note that there will be as many fields as headers, regardless of how
	 * long is the provided string array.
	 * @param values Array containing the pseudo-table value contents.
	 */
	public final void setValues(String[] values) {
		for (int i = 0; i < gridHeaders.length; i++) { // nValues = nHeaders
			gridValues[i] = new JTextField(colsValue);
			gridValues[i].setEditable(false);
			if (i < values.length)
				gridValues[i].setText(values[i]);
		}
	}
	
	// Nuked out a lot of accessors and mutators which would compromise implementation changes.
	// utilities
	
	/**
	 * Extracts the value fields' contents in a convenient format for further
	 * manipulation.
	 * @return Contents of value fields.
	 */
	public String[] extractValues() {
		String[] extract = new String[gridValues.length];
		for (int i = 0; i < gridValues.length; i++)
			extract[i] = gridValues[i].getText();
		return extract;
	}
	
	/**
	 * Updates the contents of already created headers.
	 * @param headers Array containing the pseudo-table header contents.
	 */
	public void updateHeaders(String[] headers) {
		for (int i = 0; i < gridHeaders.length; i++) {
			if (i < headers.length)
				gridHeaders[i].setText(headers[i]);
		}
	}
	
	/**
	 * Updates the contents of already created value fields.
	 * @param values Array containing the pseudo-table value field contents.
	 */
	public void updateValues(String[] values) {
		for (int i = 0; i < gridHeaders.length; i++) {
			if (i < values.length)
				gridValues[i].setText(values[i]);
		}
	}
}
